


def log_metrics(logger, model, outmode='starting_uni'):
    
    
    G = len([g.id for g in model.genes])
    R = len([r.id for r in model.reactions if len(set([m.id.rsplit('_',1)[-1] for m in r.metabolites]))==1])
    T = len([r.id for r in model.reactions if len(set([m.id.rsplit('_',1)[-1] for m in r.metabolites]))!=1])
    M = len([m.id for m in model.metabolites])
    uM = len(set([m.id.rsplit('_',1)[0] for m in model.metabolites]))
    gr = len([gr.id for gr in model.groups])
    bP = len([m.id for m in model.reactions.get_by_id('Biomass').reactants])

    
    if   outmode == 'starting_uni':
        logger.info(f"Starting universe: [oG: {G}, R: {R}, T: {T}, uM: {uM}, bP: {bP}]")
    elif outmode == 'uni_features':
        logger.info(f"Universe features: [oG: {G}, R: {R}, T: {T}, uM: {uM}, bP: {bP}]")
    elif outmode == 'recon_model':
        logger.info(f"Resulting model: [G: {G}, R: {R}, T: {T}, uM: {uM}, bP: {bP}]")
    elif outmode == 'loaded_model':
        logger.info(f"Loaded model: [G: {G}, R: {R}, T: {T}, uM: {uM}, bP: {bP}]")
        
        
        
def show_contributions(logger, db):
    
    # create a counter for each author
    cnt = {author: 0 for author in db['authors']['username']}
    cnt_tot = 0
    for rid, row in db['R'].iterrows():
        for author in row['author'].split(';'):
            author = author.rstrip().strip()
            cnt[author] += 1
            cnt_tot += 1
    for rid, row in db['T'].iterrows():
        for author in row['author'].split(';'):
            author = author.rstrip().strip()
            cnt[author] += 1
            cnt_tot += 1
        
    # compute percentages:
    pct = {author: cnt[author]/cnt_tot*100 for author in cnt.keys()}
    # sort in descending order: 
    pct = dict(sorted(pct.items(), key=lambda item: item[1], reverse=True))
    # convert to string:
    pct = {author: f'{round(pct[author],2)}%' for author in pct.keys()}
    logger.debug(f"Contributions: {pct}.")


import gempipe



def apply_medium_given_column(logger, model, medium, column, is_reference=False):
        
        
    # retrieve metadata
    description = column.iloc[0]
    doi = column.iloc[1]
    author = column.iloc[2]
    units = column.iloc[3]

    
    # convert substrates to dict
    column = column.iloc[4:]
    column = column.to_dict()

    
    # add trace elements:
    column['fe2'] = 'NL'
    column['mobd'] = 'NL'
    column['cobalt2'] = 'NL'


    # reset exchanges
    gempipe.reset_growth_env(model)    
    modeled_rids = [r.id for r in model.reactions]


    for substrate, value in column.items():

        if type(value)==float:
            continue   # empty cell, exchange will remain close

            
        # check if exchange is modeled
        if is_reference == False: 
            if f'EX_{substrate}_e' not in modeled_rids:
                logger.error(f"No exchange reaction found for substrate '{substrate}' in medium '{medium}'.")
                return 1
        else:  # external reference models might follow different standards.
            # The exr might not be present. 
            if f'EX_{substrate}_e' not in modeled_rids:
                logger.info(f"Reference has no exchange reaction for '{substrate}' in medium '{medium}': this substrate will be ignored.")
                continue


        # case "not limiting"
        value = value.strip().rstrip()
        if value == 'NL':   # non-limiting case
            model.reactions.get_by_id(f'EX_{substrate}_e').lower_bound = -1000


        # case "single value"
        elif '+-' not in value and '±' not in value:  # single number case
            value = value.replace(' ', '')  # eg "- 0.03" --> "-0.03"
            try: value = float(value)
            except: 
                logger.error(f"Invalid value found in medium '{medium}': '{substrate}' {value}.")
                return 1
            model.reactions.get_by_id(f'EX_{substrate}_e').lower_bound = value


        # case "with experimental error"
        else:  # value +- error
            if '±' in value: 
                value, error = value.split('±', 1)
            else: value, error = value.split('+-', 1)
            value = value.rstrip()
            error = error.strip()
            value = value.replace(' ', '')  # eg "- 0.03" --> "-0.03"
            try: value = float(value)
            except: 
                logger.error(f"Invalid value found in medium '{medium}': '{substrate}' {value} +- {error}.")
                return 1
            try: error = float(error)
            except: 
                logger.error(f"Invalid value found in medium '{medium}': '{substrate}' {value} +- {error}.")
                return 1
            model.reactions.get_by_id(f'EX_{substrate}_e').lower_bound = value -error
            model.reactions.get_by_id(f'EX_{substrate}_e').upper_bound = value +error

    return 0


import warnings



def remove_forced(logger, model, universe, force_removal):
    
    
    if force_removal != '-':
        removed_rids = []
        forced_rids = force_removal.split(',')
        
        modeled_rids = [r.id for r in model.reactions]
        universal_rids = [r.id for r in universe.reactions]
        
        for rid in forced_rids: 
            
            if rid not in universal_rids:
                logger.info(f"Ignoring reaction ID '{rid}' since it's not included in the universe.")
                continue
            
            if rid not in modeled_rids:
                logger.debug(f"Requested reaction ID '{rid}' was already excluded from the model.")
            else:
                removed_rids.append(rid)

        # remove collected reactions:
        with warnings.catch_warnings():
            # avoid warnings like "python3.9/site-packages/cobra/core/group.py:147: UserWarning: need to pass in a list"
            warnings.simplefilter("ignore")
            model.remove_reactions(removed_rids) # this works also with rids instaead of Rs

        logger.info(f"Reactions forcibly removed: {removed_rids}.")
        


def remove_disconnected(logger, model):
    
    to_remove = []
    for m in model.metabolites:
        if len(m.reactions) == 0:
            to_remove.append(m)
    model.remove_metabolites(to_remove)
    logger.info(f"Removed {len(to_remove)} disconnected metabolites.")
    
    



import pandas as pnd
import cobra
import gempipe


from ..commons import load_minimal_media
from ..commons import apply_medium_given_column
from ..commons import verify_growth




def grow_on_media(logger, model, expcon, media, fva):
    
    
    # if working with the universe, load the minmal media from the assets
    universe_in_parsedb = False
    if expcon == None and media == None:  
        expcon, media = load_minimal_media()
        universe_in_parsedb = True
        log_for_parsedb = []
    
    
    # check if requested
    df_G = pnd.DataFrame()
    obj_id = 'max(obj)'
    
    
    # get involved media:
    if media == '-':
        logger.info(f"No media provided: growth simulations will be skipped.")
        return df_G
    media = media.split(',')
    # at least 1 medium must exist:
    if any([i in expcon['media'].columns for i in media])==False:
        logger.error(f"None of the provided media IDs exist. Available media are {list(expcon['media'].columns)}.")
        return 1
    
    
    # prepare sheet for excel output
    df_G.loc[obj_id, 'name'] = f'Objective reaction {gempipe.get_objectives(model)}'
    for r in model.reactions:
        df_G.loc[r.id, 'name'] = r.name
        
        
    if not universe_in_parsedb: logger.info(f"Growing on {len(media)} media...")
    for medium in media: 
        if medium not in expcon['media'].columns:
            logger.info(f"Medium '{medium}' does not exists and will be ignored.")
            continue
        # create dedicated column con the excel output
        df_G[f'{medium}'] = '/'
        logger.debug(f"Growing on medium '{medium}'...")
                    
        
        # apply medium
        response = apply_medium_given_column(logger, model, medium, expcon['media'][medium])
        if response == 1: return 1

    
        # perform FBA
        res_fba = verify_growth(model, boolean=False)
        df_G.loc[obj_id, f'{medium}'] = res_fba
        if universe_in_parsedb:
            log_for_parsedb.append(f"'{medium}': {res_fba}")
        
        
        # perform FVA if requested:
        if fva and (res_fba not in [0, 'infeasible']):
            logger.debug(f"FVA on medium '{medium}'...")
            df_fva = cobra.flux_analysis.flux_variability_analysis(model, model.reactions, fraction_of_optimum=0.9)
            for rid, row in df_fva.iterrows():
                df_G.loc[rid, f'{medium}'] = f"({round(row['minimum'], 3)}, {round(row['maximum'], 3)})"

    
    if universe_in_parsedb:
        logger.info(f"Growth on minimal media: {'; '.join(log_for_parsedb)}.")
    return df_G


import pandas as pnd



def write_excel_model(model, filepath, df_E, df_B, df_P, df_S):
    
    df_M = []
    df_R = []
    df_T = []
    df_A = []
    
    
    # format df_E:  # biomass precursors biosynthesis
    if df_E is not None:
        df_E.insert(0, 'mid', '')  # new columns as first
        df_E['mid'] = df_E.index
        df_E = df_E.reset_index(drop=True)
    
    # format df_B:  # biomass assembly
    if df_B is not None:
        df_B.insert(0, 'mid', '')  # new columns as first
        df_B['mid'] = df_B.index
        df_B = df_B.reset_index(drop=True)
    
    # format df_P:  phenotype screening (Biolog(R))
    if df_P is not None:
        df_P.insert(0, 'plate:well', '')  # new columns as first
        df_P['plate:well'] = df_P.index
        df_P = df_P.reset_index(drop=True)
        
    # format df_S:  metabolite synthesis
    if df_S is not None:
        df_S.insert(0, 'mid', '')  # new columns as first
        df_S['mid'] = df_S.index
        df_S = df_S.reset_index(drop=True)
            
        
        
    for m in model.metabolites: 
        
        # get kc codes: 
        if 'kegg.compound' not in m.annotation.keys():  kc_ids = ''
        else:  
            kc_ids = m.annotation['kegg.compound']
            if type(kc_ids) == str: kc_ids = [kc_ids]
            kc_ids = '; '.join([i for i in kc_ids if i!='CXXXXX'])
        
        df_M.append({'mid': m.id, 'formula': m.formula, 'charge': m.charge, 'kc': kc_ids, 'name': m.name})

        
        
    for r in model.reactions:
        
        
        # handle artificial reactions
        if r.id == 'Biomass':
            df_A.append({'rid': r.id, 'rstring': r.reaction, 'type': 'biomass', 'name': r.name})
            
            
        elif len(r.metabolites) == 1:
            if len(r.metabolites)==1 and list(r.metabolites)[0].id.rsplit('_',1)[-1] == 'e': 
                df_A.append({'rid': r.id, 'rstring': r.reaction, 'type': 'exchange', 'name': r.name})
            elif r.lower_bound < 0 and r.upper_bound > 0:
                df_A.append({'rid': r.id, 'rstring': r.reaction, 'type': 'sink', 'name': r.name})
            elif r.lower_bound == 0 and r.upper_bound > 0:
                df_A.append({'rid': r.id, 'rstring': r.reaction, 'type': 'demand', 'name': r.name})
          
        
        else: # more than 1 metabolite involved
            
            # get kr codes: 
            if 'kegg.reaction' not in r.annotation.keys():  kr_ids = ''
            else:  
                kr_ids = r.annotation['kegg.reaction']
                if type(kr_ids) == str: kr_ids = [kr_ids]
                kr_ids = '; '.join([i for i in kr_ids if i!='RXXXXX'])

            # introduce reaction in the correct table: 
            r_dict = {'rid': r.id, 'rstring': r.reaction, 'kr': kr_ids, 'gpr': r.gene_reaction_rule, 'name': r.name}
            if len(set([m.id.rsplit('_',1)[-1] for m in r.metabolites])) == 1:
                df_R.append(r_dict)
            else: df_T.append(r_dict)
    
    
    
    df_M = pnd.DataFrame.from_records(df_M)
    df_R = pnd.DataFrame.from_records(df_R)
    df_T = pnd.DataFrame.from_records(df_T)
    df_A = pnd.DataFrame.from_records(df_A)
    with pnd.ExcelWriter(filepath) as writer:
        df_M.to_excel(writer, sheet_name='Metabolites', index=False)
        df_R.to_excel(writer, sheet_name='Reactions', index=False)
        df_T.to_excel(writer, sheet_name='Transporters', index=False)
        df_A.to_excel(writer, sheet_name='Artificials', index=False)
        if df_E is not None and len(df_E)!=0: df_E.to_excel(writer, sheet_name='Precursors', index=False)
        if df_B is not None: df_B.to_excel(writer, sheet_name='Biomass', index=False)
        if df_P is not None and len(df_P)!=0: df_P.to_excel(writer, sheet_name='Biolog®', index=False)
        if df_S is not None and len(df_S)!=0: df_S.to_excel(writer, sheet_name='Biosynth', index=False)


import pandas as pnd
import gempipe


from ..commons import load_minimal_media
from ..commons import adjust_biomass_equation
from ..commons import get_biomass_dict
from ..commons import apply_medium_given_column



def precursors_on_media(logger, model, universe, expcon, media, varprec, precursors):
    
    
    # if working with the universe, load the minmal media from the assets
    if expcon == None and media == None and varprec == None:  
        expcon, media = load_minimal_media()
        varprec = adjust_biomass_equation(logger, model, universe, 1.0)
    
    
    # get biomass precursors to check (all, even if not strain-specific): 
    biomass_mids = []
    for value in get_biomass_dict().values(): biomass_mids = biomass_mids + value
    
    
    # check if requested
    df_E = pnd.DataFrame()
    if not precursors:
        return df_E
    
    
    # prepare sheet for excel output
    df_E['name'] = None
    df_E['cond'] = None
    df_E['coeff'] = None
    for mid in biomass_mids:
        df_E.loc[mid, 'name'] = universe.metabolites.get_by_id(mid).name
        df_E.loc[mid, 'coeff'] = 0.01
        if mid in varprec.keys():
            df_E.loc[mid, 'cond'] = f'conditional'
            if mid not in [m.id for m in  model.reactions.Biomass.reactants]:  
                df_E.loc[mid, 'coeff'] = 'removed'
        else:
            df_E.loc[mid, 'cond'] = '/'

        
    # get involved media:
    if media == '-':
        logger.info(f"No media provided: blocked biomass precursors check will be skipped.")
        return df_E
    media = media.split(',')
    # at least 1 medium must exist:
    if any([i in expcon['media'].columns for i in media])==False:
        logger.error(f"None of the provided media IDs exist. Available media are {list(expcon['media'].columns)}.")
        return 1
            
    
    logger.info(f"Checking blocked biomass precursors on {len(media)} media...")
    for medium in media:
        if medium not in expcon['media'].columns:
            logger.info(f"Medium '{medium}' does not exists and will be ignored.")
            continue
        # create dedicated column on the excel output
        df_E[f'{medium}'] = None
        logger.debug(f"Checking blocked biomass precursors on medium '{medium}'...")
            
            
        # apply medium :
        response = apply_medium_given_column(logger, model, medium, expcon['media'][medium])
        if response == 1: return 1
    
            
        # check separately for each biomass precursor:
        for mid in biomass_mids:
            # skip conditional precursors: 
            if mid in [m.id for m in model.reactions.Biomass.reactants]:
                
            
                response, flux, status = gempipe.can_synth(model, mid=mid)
                if response == True:
                    df_E.loc[mid, f'{medium}'] = '/'
                else:
                    df_E.loc[mid, f'{medium}'] = 'blocked'
    
    
    return df_E



import os
import pickle


import cobra
import gempipe


from ..commons import force_id_on_sbml
from ..commons import log_metrics
from ..commons import get_db
from ..commons import introduce_biomass
from ..commons import write_excel_model
from ..commons import show_contributions

from .introduce import introduce_metabolites
from .introduce import introduce_reactions
from .introduce import introduce_transporters
from .introduce import introduce_sinks_demands

from .annotation import translate_annotate_genes
from .annotation import set_up_groups

from .completeness import check_completeness

from ..runsims.simplegrowth import grow_on_media
from ..runsims.precursors import precursors_on_media
from ..runsims.biosynth import biosynthesis_on_media




def main(args, logger):
    
    
    # adjust out folder path                      
    while args.outdir.endswith('/'):              
        args.outdir = args.outdir[:-1]
    os.makedirs(f'{args.outdir}/', exist_ok=True)
    
    
    # check compatibility of input parameters
    if args.progress==False and args.module==True: 
        logger.error(f"You cannot ask --module without --progress (see --help).")
        return 1
    if args.progress==False and args.focus!='-':
        logger.error(f"You cannot ask --focus without --progress (see --help).")
        return 1
    if args.progress==False and args.zeroes==True:
        logger.error(f"You cannot ask --zeroes without --progress (see --help).")
        return 1
    
    
    # check 'goodbefore' and 'onlyauthor' params
    if args.goodbefore == '-' and args.onlyauthor != '-':
        logger.error(f"--onlyauthor must be used in conjunction with --goodbefore.")
        return 1
    if   args.goodbefore == '-': args.goodbefore = [None, None, None]
    elif len(args.goodbefore.split('-')) != 3: 
        logger.error(f"Invalid syntax detected for --goodbefore.")
        return 1
    else:
        args.goodbefore = args.goodbefore.split('-')
        if args.goodbefore[0] == 'None': args.goodbefore[0] = None
        if args.goodbefore[1] == 'None': args.goodbefore[1] = None
        if args.goodbefore[2] == 'None': args.goodbefore[2] = None
    if args.onlyauthor == '-': args.onlyauthor = None
    
    
    # check and extract the required 'gsrap.maps' file
    if os.path.exists(f'{args.inmaps}') == False:
        logger.error(f"File 'gsrap.maps' not found at {args.inmaps}.")
        return 1
    try: 
        with open(f'{args.inmaps}', 'rb') as f:
            inmaps = pickle.load(f)  
    except: 
        logger.error(f"Provided file {args.inmaps} has an incorrect format.")
        return 1
    idcollection_dict = inmaps['idcollection_dict']
    summary_dict = inmaps['summary_dict']
    keggc_to_bigg = inmaps['keggc_to_bigg']
    keggr_to_bigg = inmaps['keggr_to_bigg']
    
    
    # download database and check structure
    db = get_db(logger)
    if type(db)==int: return 1

    # show simple statistics of contributions
    show_contributions(logger, db)
                                    
        
    
    ###### RECONSTRUCTION
    # create the model
    universe = cobra.Model('newuni')   
    
    # introduce M / R / T
    universe = introduce_metabolites(logger, db, universe, idcollection_dict, keggc_to_bigg, args.goodbefore[0], args.onlyauthor)
    if type(universe)==int: return 1
    universe = introduce_reactions(logger, db, universe, idcollection_dict, keggr_to_bigg, args.goodbefore[1], args.onlyauthor)
    if type(universe)==int: return 1
    universe = introduce_transporters(logger, db, universe, idcollection_dict, keggr_to_bigg, args.goodbefore[2], args.onlyauthor)
    if type(universe)==int: return 1

    # introduce sinks / demands (exchanges where included during T)
    universe = introduce_sinks_demands(logger, universe)
    if type(universe)==int: return 1

    # introducce biomass
    universe = introduce_biomass(logger, db, universe)
    if type(universe)==int: return 1



    ###### ANNOTATION
    # translate Gs to symbols and annotate them (EC, COG, GO, ...)
    universe = translate_annotate_genes(logger, universe, idcollection_dict)
    if type(universe)==int: return 1

    # introduce collectionas (groups of Rs as maps/modules)
    universe = set_up_groups(logger, universe, idcollection_dict)
    if type(universe)==int: return 1

    
    
    ###### CHECKS
    # check universe completness
    response = check_completeness(logger, universe, args.progress, args.module, args.focus, args.eggnog, args.zeroes, idcollection_dict, summary_dict)
    if response==1: return 1

    # check growth on minmal media
    df_G = grow_on_media(logger, universe, None, None, '-')
    if type(df_G)==int: return 1
    
    # check blocked biomass precursors
    df_E = precursors_on_media(logger, universe, universe, None, None, None, args.precursors)
    if type(df_E)==int: return 1

    # check blocked metabolites / dead-ends
    df_S = biosynthesis_on_media(logger, universe, None, None, args.biosynth)
    if type(df_S)==int: return 1


    
    ###### POLISHING
    # reset growth environment befor saving the model
    gempipe.reset_growth_env(universe)



    # output the universe
    cobra.io.save_json_model(universe, f'{args.outdir}/newuni.json')
    logger.info(f"'{args.outdir}/newuni.json' created!")
    cobra.io.write_sbml_model(universe, f'{args.outdir}/newuni.xml')   # groups are saved only to SBML 
    logger.info(f"'{args.outdir}/newuni.xml' created!")
    force_id_on_sbml(f'{args.outdir}/newuni.xml', 'newuni')   # force introduction of the 'id=""' field
    write_excel_model(universe, f'{args.outdir}/newuni.parsedb.xlsx', df_E, None, None, df_S)  
    logger.info(f"'{args.outdir}/newuni.parsedb.xlsx' created!")
    log_metrics(logger, universe, outmode='uni_features')


      
    return 0


import warnings
import logging



def get_optthr():
    
    return 0.001   # optimization threshold



def fba_no_warnings(model): 
    
    # Ignore eventual "UserWarning: Solver status is 'infeasible'."
    with warnings.catch_warnings():  # temporarily suppress warnings for this block
        warnings.simplefilter("ignore")  # ignore all warnings
        
        # disable warnings
        cobra_logger = logging.getLogger("cobra.util.solver")
        old_level = cobra_logger.level
        cobra_logger.setLevel(logging.ERROR)   

        # perform FBA: 
        res = model.optimize()
        obj_value = res.objective_value
        status = res.status

        # restore original behaviour: 
        cobra_logger.setLevel(old_level)   

        return res, obj_value, status
    
    
    
def verify_growth(model, boolean=True):
            
    
    res, obj_value, status = fba_no_warnings(model)
    if boolean:
        if obj_value < get_optthr() or status=='infeasible':
            return False
        else: return True
    else:
        if status =='infeasible':
            return 'infeasible'
        elif obj_value < get_optthr():
            return 0
        else:
            return round(obj_value, 3)
        
        
        









import cobra



def get_biomass_dict():
    
    
    # Note: universal and conditional precursors have been defined in Xavier2017 .
    # Here presented in the same order of Xavier2017 .
    # Xavier2017: 10.1016/j.ymben.2016.12.002.
    biomass_dict = {
        ##### UNIVERSAL #####
        'ribo_nucleotides': [
            'atp_c', 'ctp_c', 'gtp_c', 'utp_c'
        ],
        'deoxyribo_nucleotides': [
            'datp_c', 'dctp_c', 'dgtp_c', 'dttp_c'
        ],
        'amino_acids': [
            'ala__L_c', 'arg__L_c', 'asn__L_c', 'asp__L_c', 'cys__L_c', 
            'gln__L_c', 'glu__L_c', 'gly_c',    'his__L_c', 'ile__L_c', 
            'leu__L_c', 'lys__L_c', 'met__L_c', 'ser__L_c', 'pro__L_c', 
            'thr__L_c', 'trp__L_c', 'tyr__L_c', 'val__L_c', 'phe__L_c'
        ],
        'cofactors_uni': [
            'nad_c',    # B3: Nicotinamide -adenine dinucleotide phosphate
            'nadp_c',   # B3: Nicotinamide -adenine dinucleotide phosphate
            'coa_c',    # B5: Coenzyme A  (dependant on 'pnto__R_c')
            'fad_c',    # B2: Flavin adenine dinucleotide
            'fmn_c',    # B2: Flavin mononucleotide
            'ribflv_c', # B2: ribovlavin. Non-active form acording to Xavier2017 but anyway included.
            #'f4200_c', # B2: included by Xavier2017 in 'universal' but the description seems conditional.
            'thf_c',    # B9: tetrahydrofolate 
            '10fthf_c', # B9: 10-Formyltetrahydrofolate
            '5mthf_c',  # B9: 5-Methyltetrahydrofolate
            'thmpp_c',  # B1: Thiamine diphosphate
            'pydx5p_c', # B6: pyridoxal 5-phosphate
            'amet_c',   # SAM: S-adenosyl-methionine
        ],
        ##### CONDITIONAL #####
        'cofactors_con': [
            #'f4200_c', # coenzyme f420 (electron transfer in methanogens, actinobacteria , and others)
            'ptrc_c',   # Putrescine
            'spmd_c',   # Sperimidine
            'pheme_c',  # protoheme
            'mql8_c',   # menaquinol / manaquinone (mqn8_c)
            'q8h2_c',   # ubiquinol / ubiquinone (q8_c)
            # Methionaquinone
            'btn_c',    # B7: biotin
            'ACP_c',    # Acyl-carrier protein
            'adocbl_c', # B12: Adenosylcobalamin
            # Lipoate
            'uacgam_c'  # uridine diphosphate N-Acetylglucosamine (UDP-GlcNAc)
        ],
        ##### ADDED ##### (conditionals not included or lumped in Xavier2017)
        'cofactors_add': [
            'hemeO_c',  # heme-O
            'sheme_c',  # siroheme
            'moco_c',   # molybdenum cofactor
            'phllqol_c',# phylloquinol / phylloquinone (phllqne_c)
            'gthrd_c',  # glutathione (reduced)
            'br_c',     # bacterioruberin
        ],
        'lipids': [
            'pe120_c', # phosphatidyl-ethanolamine (12:0;12:0)
            'pg120_c', # phosphatidyl-glycerol (12:0;12:0)
            'clpn120_c', # cardiolipin (12:0;12:0;12:0;12:0)
            # 1-lysyl phosphatidylglycerol (plantarum)
        ],
        'membrane_wall': [
            'peptidoSTA_c', # peptidoglycan (dependant on 'udcpdp_c')
            'WTAgg40r_20n_20a_P_c', # teichoic acids
            'WTArg40r_20g_20a_P_c', # teichoic acids
            'WTAg40g_20g_20a_P_c', # teichoic acids
            'LTAgg40g_20n_20a_c', # lipoteichoic acids
            'LTAga40g_20t_20a_c', # lipoteichoic acids            
            # capsular polysaccharides
            # kdo_lipid_A
        ],
        'energy_stock': [
            # glycogen
            # starch
            'phb_c', # PHA / PHB
        ]
    }
    return biomass_dict



def introduce_biomass(logger, db, model): 
    
    
    biomass_dict = get_biomass_dict()
    
    
    rstring =           f'0.01 {" + 0.01 ".join(biomass_dict["ribo_nucleotides"])}'
    rstring = rstring + f' + 0.01 {" + 0.01 ".join(biomass_dict["deoxyribo_nucleotides"])}'
    rstring = rstring + f' + 0.01 {" + 0.01 ".join(biomass_dict["amino_acids"])}'
    rstring = rstring + f' + 0.01 {" + 0.01 ".join(biomass_dict["cofactors_uni"])}'
    rstring = rstring + f' + 0.01 {" + 0.01 ".join(biomass_dict["cofactors_con"])}'
    rstring = rstring + f' + 0.01 {" + 0.01 ".join(biomass_dict["cofactors_add"])}'
    rstring = rstring + f' + 0.01 {" + 0.01 ".join(biomass_dict["lipids"])}'
    rstring = rstring + f' + 0.01 {" + 0.01 ".join(biomass_dict["membrane_wall"])}'
    rstring = rstring + f' + 0.01 {" + 0.01 ".join(biomass_dict["energy_stock"])}'
    rstring = rstring + f' + 0.01 atp_c + 0.01 h2o_c --> 0.01 adp_c + 0.01 h_c + 0.01 pi_c'
    
    
    r = cobra.Reaction('Biomass')
    model.add_reactions([r])
    r = model.reactions.get_by_id('Biomass')
    r.name = 'Biomass assembly reaction'
    r.build_reaction_from_string(rstring)
    
    
    # set as objective:
    model.objective = 'Biomass'
    
    
    return model



def adjust_biomass_equation(logger, model, universe, conditional_threshold):
    
    
    # Note: universal and conditional precursors have been defined in 10.1016/j.ymben.2016.12.002 .
    precursor_to_pathway = {  # listing alternative biosynthetic pathways
        
        ##### cofactors_con #####
        #'f4200_c', # coenzyme f420 (electron transfer in methanogens, actinobacteria , and others)
        #'ptrc_c',   # Putrescine
        #'spmd_c',   # Sperimidine
        'pheme_c': ['M00868', 'M00121', 'M00926'],   # protoheme (heme)
        'mql8_c': ['M00116', 'M00930', 'M00931'],  # menaquinol
        'q8h2_c': ['M00117', 'M00989', 'M00128'],  # ubiquinol
        # Methionaquinone
        #'btn_c',    # B7: biotin
        #'ACP_c',    # Acyl-carrier protein
        'adocbl_c': ['M00122'],   # vitamin B12 (cobolamin)
        # Lipoate
        #'uacgam_c'  # uridine diphosphate N-Acetylglucosamine (UDP-GlcNAc)
        
        ##### cofactors_add #####
        'hemeO_c': ['gr_HemeO'],  # heme-O
        'sheme_c': ['M00846'],   # siroheme
        'moco_c': ['M00880'],  # molybdenum cofactor
        'phllqol_c': ['M00932'],  # phylloquinol
        'gthrd_c': ['M00118'],  # Reduced glutathione
        'br_c': ['gr_br'],  # bacterioruberin
        
        ##### lipids #####
        #'pe120_c', # phosphatidyl-ethanolamine (12:0;12:0)
        #'pg120_c', # phosphatidyl-glycerol (12:0;12:0)
        #'clpn120_c', # cardiolipin (12:0;12:0;12:0;12:0)
        
        ##### membrane_wall #####
        # 1-lysyl phosphatidylglycerol (plantarum)
        'peptidoSTA_c': ['gr_ptdSTA'], # peptidoglycan (dependant on 'udcpdp_c')
        'WTAgg40r_20n_20a_P_c': ['gr_WTA3'], # teichoic acids
        'WTArg40r_20g_20a_P_c': ['gr_WTA2'], # teichoic acids
        'WTAg40g_20g_20a_P_c':  ['gr_WTA1'], # teichoic acids
        'LTAgg40g_20n_20a_c':  ['gr_LTA1'], # lipoteichoic acids
        'LTAga40g_20t_20a_c':  ['gr_LTA2'], # lipoteichoic acids            
        # capsular polysaccharides
        # kdo_lipid_A
        
        ##### energy_stock #####
        'phb_c': ['gr_PHA1'],  # PHA / PHB
        
    }
    modeled_rids = [r.id for r in model.reactions]
    
    
    cnt_removed = 0
    varprec = {}  # dictionary of variable biomass precursors
    for precursor, pathways in precursor_to_pathway.items(): 
        
        pathway_to_absence = {}
        pathway_to_compstring = {}   # completeness string
        for pathway in pathways:   # more pathways might lead to the same precursor
            # initialize counters:
            cnt_members_tot = 0
            cnt_members_present = 0

            
            if pathway not in [gr.id for gr in universe.groups]:
                continue   # still missing from the universe
                
            for member in universe.groups.get_by_id(pathway).members:
                cnt_members_tot += 1    
                if member.id in modeled_rids:
                    cnt_members_present += 1
            # populate dicts:
            pathway_to_absence[pathway] = (cnt_members_present / cnt_members_tot) < conditional_threshold
            pathway_to_compstring[pathway] = f'{pathway}: {cnt_members_present}/{cnt_members_tot}'
            

        varprec[precursor] = '; '.join(list(pathway_to_compstring.values()))
        if all(list(pathway_to_absence.values())):
            cnt_removed += 1
            logger.debug(f"Biomass precursor '{precursor}' seems not required ({varprec[precursor]}).")
            # add metabolites to the right side (they will disappear if the balance if 0)
            model.reactions.Biomass.add_metabolites({precursor: 0.01})

       
    if model.id != universe.id:   # avoid printing when testing blocked biomass precusors of the universe.
        logger.info(f"Removed {cnt_removed} biomass precursors.")
        
        
    return varprec





def get_deprecated_kos():
    deprecated_kos = [
        'K11189',  # should be K02784
    ]
    return deprecated_kos



def get_rids_with_mancheck_gpr():
    rids_mancheck_gpr = [  # reactions with manually checked GPRs
        'SUCD1', 'ALKP', 'PFK_3', 'TCMPTS', 'PPA', 'APSR',
        'NADHPO', 'ACOD', '5DOAN', '5DRBK', '5DRBPI', '5DRBUPD',
        'FNPOR', 'THZPSN', 'PPTGP', 'ALATA_D2', 'XYLUR1'
    ]
    return rids_mancheck_gpr



def get_manual_sinks():
    
    return ['apoACP', 'aponit', 'apocarb', 'thioca', 'THI5p_b', 'cyE']



def get_manual_demands():
    
    return ['scp', 'amob', 'dialurate', 'THI5p_a', 'partmass' ]



def get_custom_groups():
    
    
    return {
        'gr_ptdSTA': ['UAMAGLL', 'UMNS', 'UPPNAPT', 'UAGPT2', 'UAGPGAL', 'UAGPN6GT', 'UAAGGTLGAGT', 'UAAGGTLG3AGT', 'PPTGP'],
        'gr_HemeO': ['HEMEOS'],
        'gr_WTA1': ['ACGAMT', 'UNDBD', 'WTAGPT', 'WTAGPP', 'WTAUGLCT2', 'WTAALAT3', 'WTAPL3'],
        'gr_WTA2': ['ACGAMT', 'UNDBD', 'WTAGPT', 'WTARBT2', 'WTARPP2', 'WTAUGLCT', 'WTAALAT2', 'WTAPL2'],
        'gr_WTA3': ['ACGAMT', 'UNDBD', 'WTAGPT', 'WTAGPT2', 'WTARPP', 'WTAGLCNACT', 'WTAALAT', 'WTAPL'],
        'gr_LTA1': ['UGDIAT', 'UGLDIAT', 'GGGDAGF2', 'LIPOPO2', 'LTANACT', 'LTAALAT2'],
        'gr_LTA2': ['UGDIAT', 'UGADIAT', 'GGGDAGF', 'LIPOPO', 'LTAGAT', 'LTAALAT'],
        'gr_br': ['LYEH1', 'HIPCD1', 'LYEH2', 'HIPCD2', 'BHBRH1', 'BHBRH2'],
        'gr_PHA1': ['ACACT1r', 'AACOAR_syn', 'PHBS_syn_1', 'PHBDEP_1'],    # PHA from glycolyis
    }


import pandas as pnd
import gempipe


from ..commons import get_biomass_dict
from ..commons import apply_medium_given_column
from ..commons import verify_growth

from .gapfillutils import get_repository_nogenes
from .gapfillutils import import_from_universe



def gapfill_on_media(logger, model, universe, expcon, media, varprec, exclude_orphans):
    
    
    
    # get biomass precursors to gap-fill: 
    biomass_mids = []
    for value in get_biomass_dict().values(): biomass_mids = biomass_mids + value
    
    
    # check if requested (always)
    df_B = pnd.DataFrame()
    
    
    # prepare sheet for excel output
    df_B['name'] = None
    df_B['cond'] = None
    df_B['coeff'] = None
    for mid in biomass_mids:
        df_B.loc[mid, 'name'] = universe.metabolites.get_by_id(mid).name
        df_B.loc[mid, 'coeff'] = 0.01
        if mid in varprec.keys():
            df_B.loc[mid, 'cond'] = f'{varprec[mid]}'
            if mid not in [m.id for m in  model.reactions.Biomass.reactants]:  
                df_B.loc[mid, 'coeff'] = 'removed'
        else:  # biomass precursor is NOT variable (it's mandatory)
            df_B.loc[mid, 'cond'] = '/'
            
    
    # get involved media:
    if media == '-':
        logger.info(f"No media provided: gap-filling will be skipped.")
        return df_B
    media = media.split(',')
    # at least 1 medium must exist:
    if any([i in expcon['media'].columns for i in media])==False:
        logger.error(f"None of the provided media IDs exist. Available media are {list(expcon['media'].columns)}.")
        return 1
        
        
    # get the repository of reactions:
    repository_nogenes = get_repository_nogenes(logger, universe, exclude_orphans)
    
    
    logger.info(f"Gap-filling for biomass on {len(media)} media...")
    for medium in media: 
        if medium not in expcon['media'].columns:
            logger.info(f"Medium '{medium}' does not exists and will be ignored.")
            continue
        # create dedicated column on the excel output
        df_B[f'{medium}'] = None
        logger.debug(f"Gap-filling on medium '{medium}'...")
                    
            
        # apply medium both on universe and model:
        response = apply_medium_given_column(logger, repository_nogenes, medium, expcon['media'][medium])
        if response == 1: return 1
        if not verify_growth(repository_nogenes):
            logger.error(f"Medium '{medium}' does not support growth of universe.")
            return 1
        
        response = apply_medium_given_column(logger, model, medium, expcon['media'][medium])
        if response == 1: return 1
        if verify_growth(model):
            logger.info(f"No need to gapfill model on medium '{medium}'.")
            continue


        # launch gap-filling separately for each biomass precursor:
        for mid in biomass_mids:
            # skip conditional precursors: 
            if mid in [m.id for m in model.reactions.Biomass.reactants]:
            
                # save time if it can already be synthesized
                if gempipe.can_synth(model, mid)[0]:
                    df_B.loc[mid, f'{medium}'] = '/'
                    logger.debug(f"Gap-filled 0 reactions on medium '{medium}' for '{mid}': [].")
                    continue   # save time!

                    
                # otherwise perform the actual gap-fill:    
                #minflux = get_optthr()
                minflux = 0.1
                suggested_rids = gempipe.perform_gapfilling(model, repository_nogenes, mid, nsol=1, minflux=minflux, boost=True, verbose=False)
                if suggested_rids == None:
                    logger.error(f"The gap-filling problem seems too hard for '{mid}' on medium '{medium}': should '{mid}' be included in '{medium}'?")
                    return 1
                
                df_B.loc[mid, f'{medium}'] = '; '.join([f'+{i}' for i in suggested_rids]) 
                logger.debug(f"Gap-filled {len(suggested_rids)} reactions on medium '{medium}' for '{mid}': {suggested_rids}.")
                for rid in suggested_rids:
                    import_from_universe(model, repository_nogenes, rid, gpr='')
                
                
    return df_B




from importlib import resources


import pandas as pnd
import gempipe


from ..commons import apply_medium_given_column
from ..commons import verify_growth

from .gapfillutils import get_repository_nogenes
from .gapfillutils import import_from_universe



def edit_trans_reacs(model, exr):
    
    pass
    


def biolog_on_media(logger, model, universe, expcon, media, biolog, exclude_orphans, cnps):
    
    
    # load assets:
    official_pm_tables = {}
    for pm in ['PM1', 'PM2A', 'PM3B', 'PM4A']:
        with resources.path("gsrap.assets", f"{pm}.csv") as asset_path: 
            official_pm_tables[pm] = pnd.read_csv(asset_path, index_col=1, names=['plate','substrate','source','u1','u2','u3','u4','kc','cas'])
    
    
    # check if requested
    df_P = pnd.DataFrame()
    if biolog == '-':  # 'cnps' is != '-' as checked during main()
        return df_P
    
    
    # format starting C/N/P/S sources as EXR:
    if len(cnps.split(',')) != 4:
        logger.error("Parameter --cnps must be formatted as 4 comma-sperated metabolite IDs (order: C, N, P and S).")
        return 1
    cnps = cnps.split(',')
    # add '_e and 'EX_' where needed.
    cnps = [i + '_e' for i in cnps if i.endswith('_e')==False]
    cnps = ["EX_" + i for i in cnps if i.startswith('EX_')==False]
    cnps = {source: exr for source, exr in zip(['carbon', 'nitrogen','sulfur','phosphorus'], cnps)}
    
    
    # get involved media:
    if media == '-':
        logger.info(f"No media provided: Biolog(R)-based model curation will be skipped.")
        return df_P
    media = media.split(',')
    # at least 1 medium must exist:
    if any([i in expcon['media'].columns for i in media])==False:
        logger.error(f"None of the provided media IDs exist. Available media are {list(expcon['media'].columns)}.")
        return 1
    
    
    # get plates for this strain
    avail_plates = []
    for pm in ['PM1', 'PM2A', 'PM3B', 'PM4A']:
        if biolog in expcon[pm].columns: 
            avail_plates.append(pm)
    if avail_plates == []:
        logger.info(f"No Biolog(R) plates found for strain '{biolog}': Biolog(R)-based model curation will be skipped.")
        return df_P
    else:
        logger.info(f"Found {len(avail_plates)} Biolog(R) plates for strain '{biolog}': {sorted(avail_plates)}.")
        
        
    # get kc-to-exr dict using built-in annotations:
    kc_to_exr = {}
    for m in model.metabolites: 
        if m.id.endswith('_e') == False:
            continue
        if 'kegg.compound' not in m.annotation.keys():
            continue
        kc_ids = m.annotation['kegg.compound']
        if type(kc_ids) == str: kc_ids = [kc_ids]
        kc_ids = [i for i in kc_ids if i != 'CXXXXX']  
        for kc_id in kc_ids:
            kc_to_exr[kc_id] = f'EX_{m.id}'
        
        
    # prepare sheet for excel output
    for pm in avail_plates:
        for well, row in official_pm_tables[pm].iterrows():
            # write substrate name:
            df_P.loc[f"{pm}:{well}", 'substrate'] = row['substrate']
            
            
            # write source type:
            if pm in ['PM1', 'PM2A']: 
                df_P.loc[f"{pm}:{well}", 'source'] = 'carbon'
            elif pm == 'PM3B': 
                df_P.loc[f"{pm}:{well}", 'source'] = 'nitrogen'
            else:
                if well[0] in ['F','G','H']: df_P.loc[f"{pm}:{well}", 'source'] = 'sulfur'
                else: df_P.loc[f"{pm}:{well}", 'source'] = 'phosphorus'
            
            
            # get kc and write the correspondent exchange
            kc = row['kc']
            if type(kc)==float: 
                if row['substrate'] == 'Negative Control': 
                    df_P.loc[f"{pm}:{well}", 'exchange'] = ''   # nagative control well
                else:
                    df_P.loc[f"{pm}:{well}", 'exchange'] = 'missing KEGG codes'  # No C/D/G codes at all
            elif kc.startswith('C'):
                if kc not in kc_to_exr.keys():
                    df_P.loc[f"{pm}:{well}", 'exchange'] = f'NA ({kc})'   # kc available, but still no transporters in db.
                else:
                    df_P.loc[f"{pm}:{well}", 'exchange'] = kc_to_exr[kc]
            elif kc.startswith('D'):
                df_P.loc[f"{pm}:{well}", 'exchange'] = 'unhandled KEGG DRUG code'  # TODO manage exchanges in this case
            elif kc.startswith('G'):
                df_P.loc[f"{pm}:{well}", 'exchange'] = 'unhandled KEGG GLYCAN code'  # TODO manage exchanges in this case
            else:
                df_P.loc[f"{pm}:{well}", 'exchange'] = '???'  # there should be no other cases
    

                
    # get the repository of reactions:
    repository_nogenes = get_repository_nogenes(logger, universe, exclude_orphans)
    
    
    logger.info(f"Performing Biolog(R)-based model curation on {len(media)} media...")
    for medium in media: 
        if medium not in expcon['media'].columns:
            logger.info(f"Medium '{medium}' does not exists and will be ignored.")
            continue
        # create dedicated column on the excel output
        df_P[f'{medium}'] = None
        logger.debug(f"Performing Biolog(R)-based model curation on medium '{medium}'...")
        
        
        # apply medium both on universe and model:
        # (growth was already verified in 'gapfill_on_media')
        response = apply_medium_given_column(logger, repository_nogenes, medium, expcon['media'][medium])
        if response == 1: return 1
        response = apply_medium_given_column(logger, model, medium, expcon['media'][medium])
        if response == 1: return 1
    
    
        # apply minimal medium to the universe: 
        if 'EX_photon870_e' in model.medium.keys():
            logger.debug("Setting universe on minimal medium 'M9photo'...")
            response = apply_medium_given_column(logger, universe, 'M9photo', expcon['media']['M9photo'])
            if response == 1: return 1
        elif 'EX_o2_e' not in model.medium.keys():
            logger.debug("Setting universe on minimal medium 'M9an'...")
            response = apply_medium_given_column(logger, universe, 'M9an', expcon['media']['M9an'])
            if response == 1: return 1
        elif 'EX_o2_e' in model.medium.keys():
            logger.debug("Setting universe on minimal medium 'M9'...")
            response = apply_medium_given_column(logger, universe, 'M9', expcon['media']['M9'])
            if response == 1: return 1
        else:
            logger.debug("No adequate minimal medium found for universe.")
            return 1
        
    
    
        # iterate pm:wells
        for index, row in df_P.iterrows():
            
            
            # get needed infos
            pm, well = index.split(':')
            source = row['source']
            exr = df_P.loc[f"{pm}:{well}", 'exchange']
            if exr.startswith('EX_') == False:
                exr = None
            start_exr = cnps[source]
            experimental = expcon[pm].loc[well, biolog]
            
            
            # check if 'start_exr' is in the medium (model or universe is the same here)
            if start_exr not in model.medium.keys():
                logger.error(f"Provided starting {source} source ('{start_exr}') is not used in '{medium}' medium. Please correct the --cnps parameter.")
                return 1
            
            
            # skip if transporters for this well are not yet implemented:
            if exr == None:
                continue
        
        
            # Ok the exr is contained in this medium. Proceed with the 2 FBAs:
            performed_gapfilling = False
            suggested_rids = None
            performed_transedit = False
            removing_rids = None
            
            with universe, repository_nogenes, model:
                
                
                # adabpt the atmsphere of the minimal medium:
                if 'EX_o2_e' not in model.medium.keys(): universe.reactions.get_by_id('EX_o2_e').lower_bound = 0 
                if 'EX_co2_e' not in model.medium.keys(): universe.reactions.get_by_id('EX_co2_e').lower_bound = 0 
                
                
                # save the lower bound for the new exr
                lb = repository_nogenes.reactions.get_by_id(start_exr).lower_bound
                
                
                # 1st FBA
                # universe
                if   source=='carbon':     universe.reactions.get_by_id('EX_glc__D_e').lower_bound = 0
                elif source=='nitrogen':   universe.reactions.get_by_id('EX_nh4_e').lower_bound = 0
                elif source=='phosphorus': universe.reactions.get_by_id('EX_pi_e').lower_bound = 0
                elif source=='sulfur':     universe.reactions.get_by_id('EX_so4_e').lower_bound = 0
                fba0_unive = verify_growth(universe, boolean=False)
                # repository
                repository_nogenes.reactions.get_by_id(start_exr).lower_bound = 0
                fba0_repos = verify_growth(repository_nogenes, boolean=False)
                # model
                model.reactions.get_by_id(start_exr).lower_bound = 0
                fba0_model = verify_growth(model, boolean=False)         
                
                
                # 2nd FBA
                # universe
                universe.reactions.get_by_id(exr).lower_bound = -1000
                fba1_unive = verify_growth(universe, boolean=False)
                # repository
                repository_nogenes.reactions.get_by_id(exr).lower_bound = lb
                fba1_repos = verify_growth(repository_nogenes, boolean=False)
                # model
                model.reactions.get_by_id(exr).lower_bound = lb
                fba1_model = verify_growth(model, boolean=False)
                   
                 
                
                # universe cannot utilize this substrate: 
                if (fba1_unive == 'infeasible') \
                or (fba0_unive == 'infeasible' and fba1_unive != 'infeasible' and fba1_unive < get_optthr()) \
                or (fba0_unive != 'infeasible' and fba1_unive != 'infeasible' and fba1_unive <= fba0_unive):

                    logger.debug(f"Substrate {pm}':'{well}':'{exr}' (E: {experimental}; Um: {fba0_unive} ▶ {fba1_unive}; Rm: {fba0_repos} ▶ {fba1_repos}; Sm: {fba0_model} ▶ {fba1_model}): universe is unable. Please expand and/or correct the universe.")
                    df_P.loc[index, f'{medium}'] = f"universe unable"
                    #return 1

                # model cannot utilize this substrate on medium:
                elif (fba1_model == 'infeasible') \
                  or (fba0_model == 'infeasible' and fba1_model != 'infeasible' and fba1_model < get_optthr()) \
                  or (fba0_model != 'infeasible' and fba1_model != 'infeasible' and fba1_model <= fba0_model):

                    # mismatch FN:
                    if experimental == 1:

                        # universe cannot reproduce experimental in the medium:
                        if (fba1_repos == 'infeasible') \
                        or (fba0_repos == 'infeasible' and fba1_repos != 'infeasible' and fba1_repos < get_optthr()) \
                        or (fba0_repos != 'infeasible' and fba1_repos != 'infeasible' and fba1_repos <= fba0_repos):

                            logger.debug(f"Substrate {pm}':'{well}':'{exr}' (E: {experimental}; Um: {fba0_unive} ▶ {fba1_unive}; Rm: {fba0_repos} ▶ {fba1_repos}; Sm: {fba0_model} ▶ {fba1_model}): universe is unable in this medium. Please expand and/or correct the universe.")
                            df_P.loc[index, f'{medium}'] = f"universe unable in medium"
                            #return 1 

                        # universe can be used for gapfilling:
                        else:
                            logger.debug(f"Substrate {pm}':'{well}':'{exr}' (E: {experimental}; Um: {fba0_unive} ▶ {fba1_unive}; Rm: {fba0_repos} ▶ {fba1_repos}; Sm: {fba0_model} ▶ {fba1_model}): gap-filling for this substrate...")
                            
                            performed_gapfilling = True
                            #minflux = get_optthr() if fba1_model == 'infeasible' else fba1_model + get_optthr()
                            minflux = 0.1 if fba1_model == 'infeasible' else fba1_model + 0.1
                            suggested_rids = gempipe.perform_gapfilling(model, repository_nogenes, nsol=1, minflux=minflux, boost=False, verbose=False)
                            if suggested_rids == None:
                                logger.error(f"The gap-filling problem seems too hard for substrate {pm}':'{well}':'{exr} in medium '{medium}'.")
                                return 1

                            df_P.loc[index, f'{medium}'] = '; '.join([f'+{i}' for i in suggested_rids]) 
                            logger.debug(f"Gap-filled {len(suggested_rids)} reactions on medium '{medium}' for substrate {pm}':'{well}':'{exr}: {suggested_rids}.")

                    # it's a match! TN
                    else:
                        logger.debug(f"Substrate {pm}':'{well}':'{exr}' (E: {experimental}; Um: {fba0_unive} ▶ {fba1_unive}; Rm: {fba0_repos} ▶ {fba1_repos}; Sm: {fba0_model} ▶ {fba1_model}): TN match.")
                        df_P.loc[index, f'{medium}'] = f"/"

                # model can utilize this substrate on medium:
                else:

                    # mismatch FP
                    if experimental == 0:

                        logger.debug(f"Substrate {pm}':'{well}':'{exr}' (E: {experimental}; Um: {fba0_unive} ▶ {fba1_unive}; Rm: {fba0_repos} ▶ {fba1_repos}; Sm: {fba0_model} ▶ {fba1_model}): trying to improve accuracy of tranport reactions...")
                        df_P.loc[index, f'{medium}'] = f"TRANSPORT"
                        
                        edit_trans_reacs(model, exr)

                    # it's a match! TP
                    else:
                        logger.debug(f"Substrate {pm}':'{well}':'{exr}' (E: {experimental}; Um: {fba0_unive} ▶ {fba1_unive}; Rm: {fba0_repos} ▶ {fba1_repos}; Sm: {fba0_model} ▶ {fba1_model}): TP match.")
                        df_P.loc[index, f'{medium}'] = f"/"
            
            
            
            # apply modifications:
            if performed_gapfilling == True:
                for rid in suggested_rids:
                    import_from_universe(model, repository_nogenes, rid, gpr='')
            if performed_transedit == True:
                pass   # TODO
            

    return df_P




import warnings


import cobra



def import_from_universe(model, universe, rid, bounds=None, gpr=None):

    
    # get the universal reaction
    ru = universe.reactions.get_by_id(rid)
    
    # create a new empty reaction
    r = cobra.Reaction(rid)
    model.add_reactions([r])
    r = model.reactions.get_by_id(rid)
    
    # copy the name
    r.name = ru.name
    
    # build string all universal metabolites are still there
    # (remove__disconnected is called later):
    r.build_reaction_from_string(ru.reaction)
    
    # set bounds
    if bounds != None: r.bounds = bounds
    else: r.bounds = ru.bounds
        
    # set GPR
    if gpr != None:
        r.gene_reaction_rule = gpr
    else:
        r.gene_reaction_rule = ''
    r.update_genes_from_gpr()



def include_forced(logger, model, universe, force_inclusion):
    
    
    if force_inclusion != '-':
        introduced_rids = []
        forced_rids = force_inclusion.split(',')
        
        modeled_rids = [r.id for r in model.reactions]
        universal_rids = [r.id for r in universe.reactions]
        
        for rid in forced_rids: 
            
            if rid not in universal_rids:
                logger.info(f"Ignoring reaction ID '{rid}' since it's not included in the universe.")
                continue
            
            if rid not in modeled_rids:
                import_from_universe(model, universe, rid, gpr='')
                introduced_rids.append(rid)
            else:
                logger.debug(f"Requested reaction ID '{rid}' was already included.")
                
        logger.info(f"Reactions forcibly included and orphans: {introduced_rids}.")
        
        

def get_repository_nogenes(logger, universe, exclude_orphans):
    # Provide a gene-free, evetually orphan-free  repository: reference or universe.


    # make an editable copy:
    repository_nogenes = universe.copy()
    
    
    if exclude_orphans:
            
            
        # collect orphan:
        to_remove = []
        for r in repository_nogenes.reactions: 
            if len(r.genes) == 0:
                if len(r.metabolites) != 1 and r.id != 'Biomass':   # exclude exchanges/sinks/demands/biomass
                    to_remove.append(r)
                    logger.debug(f"Removing orphan: {r.id} ({r.reaction}).")
              
            
        # remove orphan reactions:
        with warnings.catch_warnings():
            # avoid warnings like "python3.9/site-packages/cobra/core/group.py:147: UserWarning: need to pass in a list"
            warnings.simplefilter("ignore")
            repository_nogenes.remove_reactions(to_remove)
        logger.info(f"Removed {len(to_remove)} orphans before gap-filling.")


    # remove genes to avoid the "ValueError: id purP is already present in list"
    cobra.manipulation.delete.remove_genes(repository_nogenes, [g.id for g in repository_nogenes.genes], remove_reactions=False)

    
    return repository_nogenes




    
    




import cobra


from .manual import get_deprecated_kos
from .manual import get_custom_groups



def translate_annotate_genes(logger, model, idcollection_dict):
    

       
    ko_to_name = idcollection_dict['ko_to_name']
    ko_to_symbols = idcollection_dict['ko_to_symbols']
    ko_to_ecs = idcollection_dict['ko_to_ecs']
    ko_to_cogs = idcollection_dict['ko_to_cogs']
    ko_to_gos = idcollection_dict['ko_to_gos']
    
    
    # translation dicts: assign to each KO a symbol that is unique in the universe model.
    ko_to_sym = {}
    sym_to_ko = {}
    cnt = 0
    for g in model.genes:
        if g.id in ['orphan', 'spontaneous']: 
            continue
        ko = g.id
        cnt += 1
        
        if ko in get_deprecated_kos():
            # if the ko is deprecated, it was not included in 'ko_to_symbols'
            ko_to_sym[ko] = ko
            sym_to_ko[ko] = ko
            continue
            
        for symbol in ko_to_symbols[ko]:  # iterate the available symbols for this KO
            if symbol not in sym_to_ko.keys():   # take the first available (not yet used)
                ko_to_sym[ko] = symbol
                sym_to_ko[symbol] = ko
                break
        
        if cnt != len(ko_to_sym):  # no symbol was assigned (symbol was already taken by another KO)
            cnt_dups = 2
            symbol = list(ko_to_symbols[ko])[0] + f'_{cnt_dups}'   # generate a new symbol
            while cnt != len(ko_to_sym):   # until a symbol is assigned
                if symbol not in sym_to_ko.keys():   # if the new symbol fits
                    ko_to_sym[ko] = symbol
                    sym_to_ko[symbol] = ko
                cnt_dups += 1
                symbol = list(ko_to_symbols[ko])[0] + f'_{cnt_dups}'   # retry with the next one
                

                
    
    # insert annotations
    for g in model.genes:
        if g.id in ['orphan', 'spontaneous']: 
            continue
        ko = g.id
        g.annotation['ko'] = ko
        
        if ko not in get_deprecated_kos():
            # deprecated kos are missing from these dicts
            g.name = ko_to_name[ko]
            g.annotation['symbols'] = list(ko_to_symbols[ko])
            g.annotation['ec'] = list(ko_to_ecs[ko])
            g.annotation['cog'] = list(ko_to_cogs[ko])
            g.annotation['go'] = list(ko_to_gos[ko])
        
    
        
    # finally apply translations of IDs
    translation_dict = ko_to_sym
    translation_dict['orphan'] = 'orphan'
    translation_dict['spontaneous'] = 'spontaneous'
    cobra.manipulation.rename_genes(model, translation_dict)
    
    
    return model
    


def set_up_groups(logger, model, idcollection_dict):
    

       
    kr_to_maps = idcollection_dict['kr_to_maps']
    map_to_name = idcollection_dict['map_to_name']
    kr_to_mds = idcollection_dict['kr_to_mds']
    md_to_name = idcollection_dict['md_to_name']
    
    
    # define groups of available contents
    groups = {}   # mixing maps and mds
    for r in model.reactions:
        
        if 'kegg.reaction' not in r.annotation.keys():
            continue   # Biomass, exchanges, demands, sinks, transporters
        kr_ids = r.annotation['kegg.reaction']
            
        for kr_id in kr_ids:
            if kr_id == 'RXXXXX':
                continue
            
            # insert maps
            for map_id in kr_to_maps[kr_id]:
                if map_id not in groups.keys():
                    groups[map_id] = set()
                groups[map_id].add(r)
                
            # insert mds
            for md_id in kr_to_mds[kr_id]:
                if md_id not in groups.keys():
                    groups[md_id] = set()
                groups[md_id].add(r)
                
    # finally insert groups
    for group_id in groups.keys():
                
        # get group name
        if group_id.startswith('map'):
            name = map_to_name[group_id]
        if group_id.startswith('M'):
            name = md_to_name[group_id]
                    
        actual_group = cobra.core.Group(
            group_id, 
            name = name,
            members = list(groups[group_id]),
            kind = 'partonomy',
        )
        model.add_groups([actual_group])
        
        
        
    # insert custom groups:
    custom_groups = get_custom_groups()
    for group_id in custom_groups.keys():
        actual_group = cobra.core.Group(
            group_id, 
            name = group_id,
            members = [model.reactions.get_by_id(rid) for rid in custom_groups[group_id]],
            kind = 'partonomy',
        )
        model.add_groups([actual_group])
        
        
    return model






def growth_factors_on_media(logger, model, expcon, media, factors):
        
    
    # prepare biomass sheet for excel output
    df_F = pnd.DataFrame()
    
    
    # get involved media:
    if factors == False:
        return df_F
    if media == '-':
        logger.info(f"No media provided: essential genes prediction will be skipped.")
        return df_F
    media = media.split(',')
    # at least 1 medium must exist:
    if any([i in expcon['media'].columns for i in media])==False:
        logger.error(f"None of the provided media IDs exist. Available media are {list(expcon['media'].columns)}.")
        return 1
    
    
    # define rows
    df_F['mid'] = ''
    df_F['name'] = ''

    
    logger.info(f"Predicting growth factors on {len(media)} media...")
    for medium in media: 
        if medium not in expcon['media'].columns:
            logger.info(f"Medium '{medium}' does not exists and will be ignored.")
            continue
        # create dedicated column on the excel output
        df_F[f'{medium}'] = None
        logger.debug(f"Predicting growth factors on medium '{medium}'...")
        
        
        # apply medium
        response = apply_medium_given_column(logger, model, medium, expcon['media'][medium])
        if response == 1: return 1
    
    
        # verify growth:
        obj = verify_growth(model, boolean=False)
        if obj not in ['infeasible', 0]:
            df_F[f'{medium}'] = ''
        else:
            while obj in ['infeasible', 0]:
                res_dict = gempipe.sensitivity_analysis(model, scaled=False, top=1)
                for exr_id, value in res_dict.items():
                    if value < 0:
                        model.reactions.get_by_id(exr_id).lower_bound = -1000
                        obj = verify_growth(model, boolean=False)
                        df_F.loc[exr_id, f'{medium}'] = 'ADD'

         
    # populate 'mid'/'name' columns:
    for exr_id, row in df_F.iterrows():
        r = model.reactions.get_by_id(exr_id)
        m = list(r.metabolites)[0]
        df_F.loc[exr_id, 'mid'] = m.id.rsplit('_', 1)[0]
        df_F.loc[exr_id, 'name'] = m.name
    # replace index with 'mid':
    df_F = df_F.set_index('mid', drop=True)
        
        
    return df_F






def essential_genes_on_media(logger, model, expcon, media, essential):
        
    
    # prepare biomass sheet for excel output
    df_E = pnd.DataFrame()
    
    
    # get involved media:
    if essential == False:
        return df_E
    if media == '-':
        logger.info(f"No media provided: essential genes prediction will be skipped.")
        return df_E
    media = media.split(',')
    # at least 1 medium must exist:
    if any([i in expcon['media'].columns for i in media])==False:
        logger.error(f"None of the provided media IDs exist. Available media are {list(expcon['media'].columns)}.")
        return 1
    

    # define rows
    for g in model.genes:
        if g.id in ['spontaneous', 'orphan']:
            continue
        df_E.loc[f"{g.id}", 'name'] = g.name    
        df_E.loc[f"{g.id}", 'involving'] = '; '.join([r.id for r in g.reactions])
    
    
    logger.info(f"Predicting essential genes on {len(media)} media...")
    for medium in media: 
        if medium not in expcon['media'].columns:
            logger.info(f"Medium '{medium}' does not exists and will be ignored.")
            continue
        # create dedicated column on the excel output
        df_E[f'{medium}'] = None
        logger.debug(f"Predicting essential genes on medium '{medium}'...")
        
        
        # apply medium
        response = apply_medium_given_column(logger, model, medium, expcon['media'][medium])
        if response == 1: return 1
    
    
        # verify growth:
        obj = verify_growth(model, boolean=False)
        if obj in ['infeasible', 0]:
            df_E[f'{medium}'] = 'NA'
            continue
    
        
        single = cobra.flux_analysis.single_gene_deletion(model)
        for index, row in single.iterrows(): 
            gid = list(row['ids'])[0]  # they are single deletions
            if gid in ['spontaneous', 'orphan']:
                continue
                

            if row['status'] == 'infeasible': 
                df_E.loc[f"{gid}", medium] = 'True'
            elif row['growth'] < get_optthr():
                df_E.loc[f"{gid}", medium] = 'True'
            else: 
                df_E.loc[f"{gid}", medium] = round(row['growth'], 3)
    
    return df_E





def clean_formula(formula):
    
    # avoid confusion with 'C':
    formula = formula.replace('Ca', '').replace('Co', '').replace('Cu', '').replace('Cd', '').replace('Cr', '').replace('Cs', '').replace('Cl', '')   
    # avoid confusion with 'N':
    formula = formula.replace('Na', '').replace('Nb', '').replace('Ni', '').replace('Ne', '')
    # avoid confusion with 'P':
    formula = formula.replace('Pd', '').replace('Pt', '').replace('Pb', '').replace('Po', '')
    # avoid confusion with 'S':
    formula = formula.replace('Sc', '').replace('Si', '').replace('Sn', '').replace('Sb', '').replace('Se', '')
    
    return formula



def get_CNPS_sources(model):
    
    CNPS_sources = {'C': set(), 'N': set(), 'P': set(), 'S': set()}
    for r in model.reactions: 
        if len(r.metabolites)==1 and list(r.metabolites)[0].id.endswith('_e'):
            m = list(r.metabolites)[0]
            
            formula = m.formula
            formula = clean_formula(formula)
            
            if 'C' in formula: CNPS_sources['C'].add(r.id)
            if 'N' in formula: CNPS_sources['N'].add(r.id)
            if 'P' in formula: CNPS_sources['P'].add(r.id)
            if 'S' in formula: CNPS_sources['S'].add(r.id)
    
    return CNPS_sources


    
def get_source_name(atom):
    
    if atom=='C': return 'carbon'
    elif atom=='N': return 'nitrogen'
    elif atom=='P': return 'phosphorus'
    elif atom=='S': return 'sulfur'
    
    
    
def cnps_on_media(logger, model, expcon, media, cnps):
    
    
    # prepare biomass sheet for excel output
    df_C = pnd.DataFrame()
    
    
    # get involved media:
    if cnps == '-':
        return df_C
    if media == '-':
        logger.info(f"No media provided: alternative substrate analysis will be skipped.")
        return df_C
    media = media.split(',')
    # at least 1 medium must exist:
    if any([i in expcon['media'].columns for i in media])==False:
        logger.error(f"None of the provided media IDs exist. Available media are {list(expcon['media'].columns)}.")
        return 1
    
    
    # get main/starting C/N/P/S sources:
    if cnps == 'std': cnps = 'glc__D,nh4,pi,so4'
    startings = cnps.split(',')
    if len(startings) != 4: 
        logger.error(f"Starting sources must be 4: C, N, P, and S, in this order ({len(startings)} provided: {startings}).")
        return 1
    modeled_rids = [r.id for r in model.reactions]
    source_to_exr = {}
    for mid, source in zip(startings, ['C','N','P','S']):
        if mid.endswith('_e'): 
            mid_e = mid
        else: mid_e = mid + '_e'
        exr_id = f'EX_{mid_e}'
        if exr_id not in modeled_rids:
            logger.error(f"Expected exchange for {source} source not found (provided metabolite: '{mid}'; expected exchange: '{exr_id}').")
            return 1
        else:
            m = model.metabolites.get_by_id(mid_e)
            if source not in clean_formula(m.formula):
                logger.error(f"{source} source provided ('{mid}') does not contain {source} atoms.")
                return 1
            else:
                source_to_exr[source] = exr_id
            

    # define rows
    CNPS_sources = get_CNPS_sources(model)
    for source in ['C','N','P','S']:
        for exr_id in CNPS_sources[source]:
            m = list(model.reactions.get_by_id(exr_id).metabolites)[0]
            df_C.loc[f"[{source}] {exr_id}", 'source'] = get_source_name(source)
            df_C.loc[f"[{source}] {exr_id}", 'exchange'] = exr_id 
            df_C.loc[f"[{source}] {exr_id}", 'name'] = m.name    
    
    
    logger.info(f"Performing alternative substrate analysis on {len(media)} media...")
    for medium in media: 
        if medium not in expcon['media'].columns:
            logger.info(f"Medium '{medium}' does not exists and will be ignored.")
            continue
        # create dedicated column on the excel output
        df_C[f'{medium}'] = None
        logger.debug(f"Performing alternative substrate analysis on medium '{medium}'...")
        
        
        not_part = set()
        for source in ['C','N','P','S']:
            for exr_id in CNPS_sources[source]:
                
                # apply medium
                response = apply_medium_given_column(logger, model, medium, expcon['media'][medium])
                if response == 1: return 1
            
                
                # maybe the starting substrate is not part of the medium
                if source_to_exr[source] not in list(model.medium.keys()):
                    not_part.add(f"[{source}] {source_to_exr[source]}")
                    df_C.loc[f"[{source}] {exr_id}", medium] = f"NA"
                    continue
                
            
                model.reactions.get_by_id(source_to_exr[source]).lower_bound = -0
                obj0 = verify_growth(model, boolean=False)
                model.reactions.get_by_id(exr_id).lower_bound = -1000
                obj1 = verify_growth(model, boolean=False)
                
                
                if obj1 == 'infeasible':
                    df_C.loc[f"[{source}] {exr_id}", medium] = 'infeasible'
                elif obj0 == 'infeasible':
                    df_C.loc[f"[{source}] {exr_id}", medium] = f"{obj1} (Δ=/)"
                else:
                    df_C.loc[f"[{source}] {exr_id}", medium] = f"{obj1} (Δ={obj1-obj0})"
        
    
        # log if some starting source was not contained in this medium
        if not_part != set(): 
            logger.debug(f"Specified starting sources {not_part} were not contained in medium '{medium}'.")
    
    return df_C


import shutil
import os
import pickle


from .kdown import download_raw_txtfiles
from .kdown import create_dict_ko
from .kdown import create_dict_c
from .kdown import create_dict_r
from .kdown import create_dict_map
from .kdown import create_dict_md
from .kdown import create_idcollection_dict
from .kdown import create_summary_dict

from .mnxdown import pull_mnx_ftp
from .mnxdown import create_mnxm_to_others
from .mnxdown import create_mnxr_to_others
from .mnxdown import create_others_to_mnx
from .mnxdown import compdict_from_to



def do_kdown(logger, outdir, usecache, keeptmp): 
    
    
    logger.info(f"Respectfully retrieving metabolic information from KEGG. Raw data are being saved into '{outdir}/kdown/'. Be patient, could take a couple of days...")
    os.makedirs(f'{outdir}/kdown/', exist_ok=True)
    
    response = download_raw_txtfiles(logger, outdir, usecache)
    if type(response) == int: return 1
    else: RELEASE_kegg = response
    
    
    logger.info("Parsing downloaded KEGG information...")

    response = create_dict_ko(logger, outdir)
    if type(response) == int: return 1
    else: dict_ko = response
    
    response = create_dict_c(logger, outdir)
    if type(response) == int: return 1
    else: dict_c = response
    
    response = create_dict_r(logger, outdir)
    if type(response) == int: return 1
    else: dict_r = response
    
    response = create_dict_map(logger, outdir)
    if type(response) == int: return 1
    else: dict_map = response
    
    response = create_dict_md(logger, outdir)
    if type(response) == int: return 1
    else: dict_md = response
    
    
    # create 'idcollection_dict' and 'summary_dict' dictionaries
    idcollection_dict = create_idcollection_dict(dict_ko, dict_c, dict_r, dict_map, dict_md)
    summary_dict = create_summary_dict(dict_c, dict_r, dict_map, dict_md)
    
        
    return (RELEASE_kegg, idcollection_dict, summary_dict)



def do_mnxdown(logger, outdir, usecache, keeptmp): 
    
    
    logger.info(f"Retrieving metabolic information from MetaNetX. Raw data are being saved into '{outdir}/mnxdown/'. This should take minutes...")
    os.makedirs(f'{outdir}/mnxdown/', exist_ok=True)
    
    response = pull_mnx_ftp(logger, outdir, usecache)
    if type(response) == int: return 1
    else: RELEASE_nmx = response
    
    
    logger.info("Parsing downloaded MetaNetX information (metabolites)...")
    
    # create the 'mnx_to_others' and 'others_to_mnx' dictionaries
    mnxm_to_others = create_mnxm_to_others(logger, outdir)
    others_to_mnxm = create_others_to_mnx(logger, mnxm_to_others)
    
    logger.info("Parsing downloaded MetaNetX information (reactions)...")
    
    # create the 'mnx_to_others' and 'others_to_mnx' dictionaries
    mnxr_to_others = create_mnxr_to_others(logger, outdir)
    others_to_mnxr = create_others_to_mnx(logger, mnxr_to_others)
    
    
    mnx_assets = (mnxm_to_others, others_to_mnxm, mnxr_to_others, others_to_mnxr)
    return (RELEASE_nmx, mnx_assets)



def main(args, logger):
    
    
    # adjust out folder path                      
    while args.outdir.endswith('/'):              
        args.outdir = args.outdir[:-1]
    os.makedirs(f'{args.outdir}/', exist_ok=True)
    
    
    # KEGG
    response = do_kdown(logger, args.outdir, args.usecache, args.keeptmp)
    if type(response) == int: return 1
    else: RELEASE_kegg, idcollection_dict, summary_dict = response[0], response[1], response[2]

    
    # MetaNetX
    response = do_mnxdown(logger, args.outdir, args.usecache, args.keeptmp)
    if type(response) == int: return 1
    else: RELEASE_nmx, mnx_assets = response[0], response[1]    
    # create compact dictionaries, with just the needed information, to save space:
    keggc_to_bigg = compdict_from_to(mnx_assets, 'm', 'kegg.compound', 'bigg.metabolite')
    keggr_to_bigg = compdict_from_to(mnx_assets, 'r', 'kegg.reaction', 'bigg.reaction')


    # create 'gsrap.maps':
    with open(f'{args.outdir}/gsrap.maps', 'wb') as wb_handler:
        pickle.dump({
            'RELEASE_kegg': RELEASE_kegg, 'idcollection_dict': idcollection_dict, 'summary_dict': summary_dict,
            'RELEASE_nmx': RELEASE_nmx, 'keggc_to_bigg': keggc_to_bigg, 'keggr_to_bigg': keggr_to_bigg,
        }, wb_handler)
    logger.info(f"'{args.outdir}/gsrap.maps' created!")
    
        
    # clean temporary files:
    if not args.keeptmp:
        shutil.rmtree(f'{args.outdir}/kdown', ignore_errors=True)
        shutil.rmtree(f'{args.outdir}/mnxdown', ignore_errors=True)
        logger.info(f"Temporary raw files deleted!")
    
    
    return 0


import os
import warnings
import logging


import pandas as pnd
import cobra


from ..commons import log_metrics

    
    
def check_inputs(logger, universe, eggnog):
    
    
    # check if files exist
    if os.path.isfile(universe) == False: 
        logger.error(f"Provided --universe doesn't exist: {universe}.")
        return 1
    if os.path.isfile(eggnog) == False: 
        logger.error(f"Provided --eggnog doesn't exist: {eggnog}.")
        return 1
    
    
    # check the universe model format
    if universe.endswith('.xml'):
        universe = cobra.io.read_sbml_model(universe)
    else: 
        logger.error(f"Provided --universe must be in cobrapy-compatible SBML format (.xml extension).")
        return 1
    
    
    # log main universe metrics:
    log_metrics(logger, universe, outmode='starting_uni')
        
        
    # load eggnog annotations
    eggnog = pnd.read_csv(eggnog, sep='\t', comment='#', header=None)
    eggnog.columns = 'query	seed_ortholog	evalue	score	eggNOG_OGs	max_annot_lvl	COG_category	Description	Preferred_name	GOs	EC	KEGG_ko	KEGG_Pathway	KEGG_Module	KEGG_Reaction	KEGG_rclass	BRITE	KEGG_TC	CAZy	BiGG_Reaction	PFAMs'.split('\t')
    eggnog = eggnog.set_index('query', drop=True, verify_integrity=True)
    
    return [universe, eggnog]



def parse_eggnog(eggnog):
    
    
    # PART 1. get KO codes available
    gid_to_kos = {}
    ko_to_gids = {}
    for gid, kos in eggnog['KEGG_ko'].items():
        if kos == '-': 
            continue
            
        if gid not in gid_to_kos.keys(): 
            gid_to_kos[gid] = set()
            
        kos = kos.split(',')
        kos = [i.replace('ko:', '') for i in kos]
        for ko in kos: 
            if ko not in ko_to_gids.keys(): 
                ko_to_gids[ko] = set()
                
            # populate dictionaries
            ko_to_gids[ko].add(gid)
            gid_to_kos[gid].add(ko)

    
    return ko_to_gids, gid_to_kos



def get_modeled_kos(model):
    
    
    # get modeled KO ids:
    modeled_gid_to_ko = {}
    modeled_ko_to_gid = {}
    
    for g in model.genes:
        if g.id in ['orphan', 'spontaneous']: 
            continue
        corresponding_ko = g.annotation['ko']
        
        modeled_gid_to_ko[g.id] = corresponding_ko
        modeled_ko_to_gid[corresponding_ko] = g.id
        
    modeled_kos = list(modeled_gid_to_ko.values())
        
    return modeled_kos, modeled_gid_to_ko, modeled_ko_to_gid



def subtract_kos(logger, model, eggonog_ko_to_gids):
    
    
    modeled_kos, _, modeled_ko_to_gid = get_modeled_kos(model)
        
        
    to_remove = []  # genes to delete
    for ko in modeled_kos: 
        if ko not in eggonog_ko_to_gids.keys():
            gid_to_remove = modeled_ko_to_gid[ko]
            to_remove.append(model.genes.get_by_id(gid_to_remove))
            
    
    # remove also orphan reactions!
    to_remove.append(model.genes.get_by_id('orphan'))
    
    
    # delete marked genes!
    # trick to avoid the WARNING "cobra/core/group.py:147: UserWarning: need to pass in a list" 
    # triggered when trying to remove reactions that are included in groups. 
    with warnings.catch_warnings():  # temporarily suppress warnings for this block
        warnings.simplefilter("ignore")  # ignore all warnings
        cobra_logger = logging.getLogger("cobra.util.solver")
        old_level = cobra_logger.level
        cobra_logger.setLevel(logging.ERROR)   

        cobra.manipulation.delete.remove_genes(model, to_remove, remove_reactions=True)

        # restore original behaviour: 
        cobra_logger.setLevel(old_level)   
        
   
    logger.info(f"Found {len(model.genes)} modeled orthologs.")
    return 0



def translate_remaining_kos(logger, model, eggonog_ko_to_gids):
    
    
    _, modeled_gid_to_ko, _ = get_modeled_kos(model) 
    
    
    # iterate reactions:
    for r in model.reactions:

        gpr = r.gene_reaction_rule

        # force each gid to be surrounded by spaces: 
        gpr = ' ' + gpr.replace('(', ' ( ').replace(')', ' ) ') + ' '
        
        for gid in modeled_gid_to_ko.keys():
            if f' {gid} ' in gpr:
                
                new_gids = eggonog_ko_to_gids[modeled_gid_to_ko[gid]]
                gpr = gpr.replace(f' {gid} ', f' ({" or ".join(new_gids)}) ')       
            

        # remove spaces between parenthesis
        gpr = gpr.replace(' ( ', '(').replace(' ) ', ')')
        # remove spaces at the extremes: 
        gpr = gpr[1: -1]


        # New genes are introduced. Parethesis at the extremes are removed if not necessary. 
        r.gene_reaction_rule = gpr
        r.update_genes_from_gpr()
            
            
    # remaining old 'Cluster_'s need to removed.
    # remove if (1) hte ID starts with clusters AND (2) they are no more associated with any reaction
    to_remove = []
    for g in model.genes:
        
        if g.id in ['orphan', 'spontaneous']:
            continue
            
        if g.id in modeled_gid_to_ko.keys() and len(g.reactions)==0:
            to_remove.append(g)
            
    # warning suppression not needed here, as no reaction is actually removed.
    cobra.manipulation.delete.remove_genes(model, to_remove, remove_reactions=True)
    
        
    logger.info(f"Translated orthologs to {len(model.genes)} genes.")
    return 0
        
    

def restore_gene_annotations(logger, model, universe, eggonog_gid_to_kos):
    
    
    for g in model.genes:
        if g.id == 'spontaneous': 
            continue
            
        names = []    
        for ko in eggonog_gid_to_kos[g.id]:
            
            # get the corresponding universal gene:
            uni_g = None
            for ug in universe.genes:
                if 'ko' not in ug.annotation.keys():
                    continue
                if ug.annotation['ko']==ko:  # take the first (and only)
                    uni_g = ug
                    break
            if uni_g == None:  
                # The ko provided by eggnog-mapper is still not modeled in the universe.
                # Multiple ko are possible for each gene. Of these, only 1 could b modeled.
                continue
            
            
            # transfer annotations of this ko/universal gene:
            for key in uni_g.annotation.keys():
                if key == 'ko':
                    continue   # resulting models will loose links to kos.
                if key not in g.annotation:
                    g.annotation[key] = []
                items = uni_g.annotation[key]
                if type(items)==str:  items = [items]
                for i in items:
                    g.annotation[key].append(i)
                    
            # collect names
            names.append(uni_g.name)
        g.name = '; '.join(names)




    

    











def omission_on_media(logger, model, expcon, media, omission):
        
    
    # define metabolites for single omission experiments  
    # Note: they are all universal
    pure_mids = set([
        # 20 aminoacids
        'ala__L', 'arg__L', 'asn__L', 'asp__L', 'cys__L', 
        'gln__L', 'glu__L', 'gly', 'his__L', 'ile__L', 
        'leu__L', 'lys__L', 'met__L', 'phe__L', 'pro__L', 
        'ser__L', 'thr__L', 'trp__L', 'tyr__L', 'val__L', 
        # 5 nucleotides
        'ade', 'gua', 'csn', 'thym', 'ura'])
    modeled_rids = [r.id for r in model.reactions]
    
    
    # prepare biomass sheet for excel output
    df_O = pnd.DataFrame()
    
    
    # get involved media:
    if omission == False:
        return df_O
    if media == '-':
        logger.info(f"No media provided: single omissions will be skipped.")
        return df_O
    media = media.split(',')
    # at least 1 medium must exist:
    if any([i in expcon['media'].columns for i in media])==False:
        logger.error(f"None of the provided media IDs exist. Available media are {list(expcon['media'].columns)}.")
        return 1
    
        
    # check if the target exchange reaction exists:
    for pure_mid in list(pure_mids):
        if f'EX_{pure_mid}_e' not in modeled_rids:
            logger.debug(f"Exchange reaction 'EX_{pure_mid}_e' not found during single omissions: it will be ignored.")
            df_O.loc[pure_mid, 'name'] = 'NA'
            pure_mids = pure_mids - set([pure_mid]) 
        else:  # initialize dataframe
            m = model.metabolites.get_by_id(f'{pure_mid}_e')
            df_O.loc[pure_mid, 'name'] = m.name
    
    
    logger.info(f"Performing single omissions on {len(media)} media...")
    for medium in media: 
        if medium not in expcon['media'].columns:
            logger.info(f"Medium '{medium}' does not exists and will be ignored.")
            continue
        # create dedicated column on the excel output
        df_O[f'{medium}'] = None
        logger.debug(f"Performing single omissions on medium '{medium}'...")
        
        
        # apply medium
        response = apply_medium_given_column(logger, model, medium, expcon['media'][medium])
        if response == 1: return 1


        # set up exchange reactions:
        for pure_mid1 in pure_mids:
            for pure_mid2 in pure_mids:
                exr_id = f'EX_{pure_mid2}_e'
                
                if pure_mid2 != pure_mid1:
                    model.reactions.get_by_id(exr_id).lower_bound = -1000
                else:
                    model.reactions.get_by_id(exr_id).lower_bound = -0
                    
                    
            # perform FBA
            res_fba = verify_growth(model, boolean=False)
            df_O.loc[pure_mid1, f'{medium}'] = res_fba
               
            
    return df_O


import urllib
import os


import pandas as pnd



def pull_mnx_ftp(logger, outdir, usecache):
    
    
    def single_pull(logger, outdir, url, filename, usecache):
        if (os.path.exists(f"{outdir}/mnxdown/{filename}") and usecache) == False:
            with urllib.request.urlopen(url) as response:
                with open(f"{outdir}/mnxdown/{filename}", 'wb') as out_file:
                    out_file.write(response.read())
                    logger.debug(f"Downloaded '{outdir}/mnxdown/{filename}'!")

                    
    RELEASE_mnx = None
    single_pull(logger, outdir, 'https://www.metanetx.org/ftp/latest/CHANGELOG', 'CHANGELOG.txt', usecache) 
    with open(f"{outdir}/mnxdown/CHANGELOG.txt", "r") as f:
        RELEASE_mnx = f.read()
        
    single_pull(logger, outdir, 'https://www.metanetx.org/ftp/latest/chem_depr.tsv', 'chem_depr.tsv', usecache) 
    single_pull(logger, outdir, 'https://www.metanetx.org/ftp/latest/chem_isom.tsv', 'chem_isom.tsv', usecache) 
    single_pull(logger, outdir, 'https://www.metanetx.org/ftp/latest/chem_prop.tsv', 'chem_prop.tsv', usecache) 
    single_pull(logger, outdir, 'https://www.metanetx.org/ftp/latest/chem_xref.tsv', 'chem_xref.tsv', usecache) 
    single_pull(logger, outdir, 'https://www.metanetx.org/ftp/latest/comp_depr.tsv', 'comp_depr.tsv', usecache) 
    single_pull(logger, outdir, 'https://www.metanetx.org/ftp/latest/comp_prop.tsv', 'comp_prop.tsv', usecache) 
    single_pull(logger, outdir, 'https://www.metanetx.org/ftp/latest/comp_xref.tsv', 'comp_xref.tsv', usecache) 
    single_pull(logger, outdir, 'https://www.metanetx.org/ftp/latest/reac_depr.tsv', 'reac_depr.tsv', usecache) 
    single_pull(logger, outdir, 'https://www.metanetx.org/ftp/latest/reac_prop.tsv', 'reac_prop.tsv', usecache) 
    single_pull(logger, outdir, 'https://www.metanetx.org/ftp/latest/reac_xref.tsv', 'reac_xref.tsv', usecache) 
            
               
    return RELEASE_mnx



def remove_comment_lines(filepath):
    filtered_lines = None
    with open(filepath, 'r') as infile:
        lines = infile.readlines()
        filtered_lines = [line for line in lines if not line.lstrip().startswith('#')]
    with open(filepath, 'w') as outfile:
        outfile.writelines(filtered_lines)


        
def create_mnxr_to_others(logger, outdir):
    
    
    # Remove lines beginning with "#". Better than using "comment='#'" inside pnd.read_csv(), 
    # as some IDs in 'reac_xref.tsv' are like "kegg.reaction:R04422#1" and "kegg.reaction:R04422#2"
    remove_comment_lines(f"{outdir}/mnxdown/reac_prop.tsv")
    remove_comment_lines(f"{outdir}/mnxdown/reac_xref.tsv")

    
    # load main tables:
    header_reac_prop = ['ID', 'mnx_equation', 'reference', 'ECs', 'is_balanced', 'is_transport',]
    reac_prop = pnd.read_csv(f"{outdir}/mnxdown/reac_prop.tsv", sep='\t', header=None, names=header_reac_prop)
    reac_prop = reac_prop.set_index('ID', drop=True, verify_integrity=True)

    header_reac_xref = ['source', 'ID', 'description']
    reac_xref = pnd.read_csv(f"{outdir}/mnxdown/reac_xref.tsv", sep='\t', header=None, names=header_reac_xref)
    reac_xref = reac_xref.set_index('source', drop=True, verify_integrity=True)
    
    
    # create 'mnx_to_others':
    # 'ec' info is exclusively in 'reac_prop'
    mnx_to_others = {}
    for index, row in reac_prop.iterrows(): 
        if index.startswith('MNXR'):   
            if index not in mnx_to_others.keys(): 
                mnx_to_others[index] = {
                    'ec': set(), 
                    'kegg.reaction': set(),
                    'metacyc.reaction': set(),
                    'bigg.reaction': set(),
                    'seed.reaction': set(),
                }
                
            if pnd.isna(row['ECs']) == False:
                for ec in row['ECs'].split(';'): 
                    mnx_to_others[index]['ec'].add(ec)
                    
    
    # fill 'mnx_to_others':
    for index, row in reac_xref.iterrows():

        # get 'dbid' and 'eqid'
        dbid = index.split(':')[0]
        if dbid == 'mnx': continue
        try: eqid = index.split(':')[1].split('#')[0]  # handle cases like like "kegg.reaction:R04422#1" and "kegg.reaction:R04422#2"
        except: continue   # eg this row "MNXR02	MNXR02	PROTON import/export at model boundary"
        if eqid.startswith('R_'):
            eqid = eqid[2:]
        if row['ID'].startswith('MNXR'):

            if   dbid in ['kegg.reaction', 'keggR']:
                mnx_to_others[row['ID']]['kegg.reaction'].add(eqid)
            elif dbid in ['metacyc.reaction', 'metacycR']:
                mnx_to_others[row['ID']]['metacyc.reaction'].add(eqid)
            elif dbid in ['bigg.reaction', 'biggR']:
                mnx_to_others[row['ID']]['bigg.reaction'].add(eqid)
            elif dbid in ['seed.reaction', 'seedR']:
                mnx_to_others[row['ID']]['seed.reaction'].add(eqid)
            
            
    logger.info(f"Collected information for {len(mnx_to_others.keys())} MNX reactions.")
    return mnx_to_others



def create_mnxm_to_others(logger, outdir):
    
    
    # Remove lines beginning with "#". Better than using "comment='#'" inside pnd.read_csv(), 
    # as some IDs in 'reac_xref.tsv' are like "kegg.reaction:R04422#1" and "kegg.reaction:R04422#2"
    remove_comment_lines(f"{outdir}/mnxdown/chem_prop.tsv")
    remove_comment_lines(f"{outdir}/mnxdown/chem_xref.tsv")


    # load main tables:
    header_chem_prop = ['ID', 'name', 'reference', 'formula', 'charge', 'mass', 'InChI', 'InChIKey', 'SMILES']
    chem_prop = pnd.read_csv(f"{outdir}/mnxdown/chem_prop.tsv", sep='\t', header=None, names=header_chem_prop)
    chem_prop = chem_prop.set_index('ID', drop=True, verify_integrity=True)

    header_chem_xref = ['source', 'ID', 'description']
    chem_xref = pnd.read_csv(f"{outdir}/mnxdown/chem_xref.tsv", sep='\t', header=None, names=header_chem_xref)
    chem_xref = chem_xref.set_index('source', drop=True, verify_integrity=True)
    
    
    # create 'mnx_to_others':
    # 'formula','charge','inchikey' info is exclusively in 'chem_prop'
    mnx_to_others = {}
    for index, row in chem_prop.iterrows(): 
        if index.startswith('MNXM'):   
            if index not in mnx_to_others.keys(): 
                mnx_to_others[index] = {
                    'formula': None, 
                    'charge': None, 
                    'inchikey': None,
                    'kegg.compound': set(),
                    'kegg.drug': set(),
                    'kegg.glycan': set(),
                    'metacyc.compound': set(),
                    'bigg.metabolite': set(),
                    'seed.compound': set(), 
                }

            if pnd.isna(row['formula']) == False:
                mnx_to_others[index]['formula'] = row['formula']
            if pnd.isna(row['charge']) == False:
                mnx_to_others[index]['charge'] = float(row['charge'])
            if pnd.isna(row['InChIKey']) == False:
                inchikey = row['InChIKey']
                inchikey = inchikey.replace('InChIKey=', '')
                mnx_to_others[index]['inchikey'] = inchikey
                
                
    # fill 'mnx_to_others':
    for index, row in chem_xref.iterrows():

        # get 'dbid' and 'eqid'
        dbid = index.split(':')[0]
        if dbid == 'mnx': continue
        try: eqid = index.split(':')[1].split('#')[0]  # handle cases like like "kegg.reaction:R04422#1" and "kegg.reaction:R04422#2"
        except: continue   # eg this row "MNXM02  MNXM02  OH(-)||hydroxyde"
        if eqid.startswith('M_'):
            eqid = eqid[2:]
        if row['ID'].startswith('MNXM'):  

            if   dbid in ['kegg.compound', 'keggC']:
                mnx_to_others[row['ID']]['kegg.compound'].add(eqid)
            elif dbid in ['metacyc.drug', 'keggD']:
                mnx_to_others[row['ID']]['kegg.drug'].add(eqid)
            elif dbid in ['metacyc.glycan', 'keggG']:
                mnx_to_others[row['ID']]['kegg.glycan'].add(eqid)
            elif dbid in ['metacyc.compound', 'metacycM']:
                mnx_to_others[row['ID']]['metacyc.compound'].add(eqid)
            elif dbid in ['bigg.metabolite', 'biggM']:
                mnx_to_others[row['ID']]['bigg.metabolite'].add(eqid)
            elif dbid in ['seed.compound', 'seedM']:
                mnx_to_others[row['ID']]['seed.compound'].add(eqid)


    logger.info(f"Collected information for {len(mnx_to_others.keys())} MNX metabolites.")
    return mnx_to_others



def compdict_from_to(mnx_assets, mr='m', dbfrom=None, dbto=None ):
    
    mnxm_to_others, others_to_mnxm, mnxr_to_others, others_to_mnxr = \
        mnx_assets[0], mnx_assets[1], mnx_assets[2], mnx_assets[3]
    
    if mr == 'm':
        return {key: mnxm_to_others[value][dbto] for key, value in others_to_mnxm[dbfrom].items()}
    if mr == 'r':
        return {key: mnxr_to_others[value][dbto] for key, value in others_to_mnxr[dbfrom].items()}
    
    keggc_to_bigg = compdict_from_to(mnx_assets, 'm', 'kegg.compound', 'bigg.metabolite')
    keggr_to_bigg = compdict_from_to(mnx_assets, 'r', 'kegg.reaction', 'bigg.reaction')



def create_others_to_mnx(logger, mnx_to_others):
    # convert 'mnx_to_others' into 'mnx_to_others':

    others_to_mnx = {}
    for mnxid, others in mnx_to_others.items():
        for db in others.keys():
            if db in ['formula', 'charge', 'inchikey']:
                continue   # for the metabolite dicts
            if db not in others_to_mnx.keys():
                others_to_mnx[db] = {}
            for otherid in others[db]:
                if otherid not in others_to_mnx[db].keys():
                    others_to_mnx[db][otherid] = mnxid
                
               
    for db in others_to_mnx.keys():
        logger.info(f"Collected crossrefs for {len(others_to_mnx[db].keys())} '{db}' items.")
    return others_to_mnx


import requests
import io
from importlib import resources


import pandas as pnd



def get_db(logger):
    
    
    logger.info("Downloading the database...")
    sheet_id = "1dXJBIFjCghrdvQtxEOYlVNWAQU4mK-nqLWyDQeUZqek"
    #sheet_id = "1dCVOOnpNg7rK3iZmTDz3wybW7YrUNoClnqezT9Q5bpc" # alternative
    url = f"https://docs.google.com/spreadsheets/d/{sheet_id}/export?format=xlsx"
    response = requests.get(url)  # download the requested file
    if response.status_code == 200:
        excel_data = io.BytesIO(response.content)   # load into memory
        exceldb = pnd.ExcelFile(excel_data)
    else:
        logger.error(f"Error during download. Please contact the developer.")
        return 1
    
    
    # check table presence
    sheet_names = exceldb.sheet_names
    for i in ['T', 'R', 'M', 'authors']: 
        if i not in sheet_names:
            logger.error(f"Sheet '{i}' is missing!")
            return 1
        
        
    # load the tables
    db = {}
    db['T'] = exceldb.parse('T')
    db['R'] = exceldb.parse('R')
    db['M'] = exceldb.parse('M')
    db['authors'] = exceldb.parse('authors')
    
    
    # check table headers
    headers = {}
    headers['T'] = ['rid', 'rstring', 'kr', 'gpr_manual', 'name', 'author', 'notes']
    headers['R'] = ['rid', 'rstring', 'kr', 'gpr_manual', 'name', 'author', 'notes']
    headers['M'] = ['pure_mid', 'formula', 'charge', 'kc', 'name', 'inchikey', 'author', 'notes']
    headers['authors'] = ['username', 'first_name', 'last_name', 'role', 'mail']
    for i in db.keys(): 
        if set(db[i].columns) != set(headers[i]):
            logger.error(f"Sheet '{i}' is missing the columns {set(headers[i]) - set(db[i].columns)}.")
            return 1
        
    return db



def get_expcon(logger):
    
    
    logger.info("Downloading the experimental constraints file...")
    sheet_id = "1qGbIIipHJgYQjk3M0xDWKvnTkeInPoTeH9unDQkZPwg"
    url = f"https://docs.google.com/spreadsheets/d/{sheet_id}/export?format=xlsx"
    response = requests.get(url)  # download the requested file
    if response.status_code == 200:
        excel_data = io.BytesIO(response.content)   # load into memory
        exceldb = pnd.ExcelFile(excel_data)
    else:
        logger.error(f"Error during download. Please contact the developer.")
        return 1
    
    
    # check table presence
    sheet_names = exceldb.sheet_names
    for i in ['media', 'PM1', 'PM2A', 'PM3B', 'PM4A', 'authors']: 
        if i not in sheet_names:
            logger.error(f"Sheet '{i}' is missing!")
            return 1
        
        
    # load the tables
    expcon = {}
    expcon['media'] = exceldb.parse('media')
    expcon['PM1'] = exceldb.parse('PM1')
    expcon['PM2A'] = exceldb.parse('PM2A')
    expcon['PM3B'] = exceldb.parse('PM3B')
    expcon['PM4A'] = exceldb.parse('PM4A')
    expcon['authors'] = exceldb.parse('authors')
    
    
    # assign substrates as index
    expcon['media'].index = expcon['media'].iloc[:, 1]
    # remove first 2 useless column (empty & substrates)
    expcon['media'] = expcon['media'].iloc[:, 2:]
    
    
    for sheet in ['PM1', 'PM2A', 'PM3B', 'PM4A']:
        # assign wells as index
        expcon[sheet].index = expcon[sheet].iloc[:, 2]
        # remove first 3 useless columns
        expcon[sheet] = expcon[sheet].iloc[:, 3:]
    

    return expcon



def load_minimal_media():
    
    with resources.path("gsrap.assets", f"exp_constraints.xlsx") as asset_path: 
        expcon = pnd.read_excel(asset_path, sheet_name=None)
        
        expcon['media'].index = expcon['media'].iloc[:, 1]
        # remove first 2 useless column (empty & substrates)
        expcon['media'] = expcon['media'].iloc[:, 2:]
        
        media = ','.join(list(expcon['media'].columns))
        
        return expcon, media




